from asphalt.core.context import Context
import pytest

from asphalt.serialization.api import Serializer
from asphalt.serialization.component import SerializationComponent
from asphalt.serialization.serializers.json import JSONSerializer


@pytest.mark.asyncio
async def test_component_start():
    component = SerializationComponent(serializers={
        'json': {'encoder_options': {'allow_nan': False}},
        'msgpack': {'unpacker_options': {'encoding': 'iso-8859-1'}},
        'pickle': {'protocol': 3},
        'yaml': {'safe': False}
    })
    ctx = Context()
    await component.start(ctx)

    assert ctx.json
    assert ctx.msgpack
    assert ctx.pickle
    assert ctx.yaml


@pytest.mark.asyncio
async def test_default_config():
    component = SerializationComponent(backend='json')
    ctx = Context()
    await component.start(ctx)

    resource = await ctx.request_resource(Serializer)
    assert isinstance(resource, JSONSerializer)
    assert ctx.serializer is resource



from typeguard import check_argument_types

from ..api import Serializer

try:
    import cbor
except ImportError:  # pragma: no cover
    raise ImportError('cbor is missing -- install asphalt-serialization[cbor] to fix this')


class CBORSerializer(Serializer):
    """
    Serializes objects using CBOR (Concise Binary Object Representation).

    To use this serializer backend, the ``cbor`` library must be installed.
    A convenient way to do this is to install ``asphalt-serialization`` with the ``cbor``
    extra:

    .. code-block:: shell

        $ pip install asphalt-serialization[cbor]

    .. seealso:: `CBOR web site <http://cbor.io/>`_

    :param sort_keys: ``True`` to output dictionary items by keys when serializing
    """

    __slots__ = '_sort_keys'

    def __init__(self, sort_keys: bool=False):
        assert check_argument_types()
        self._sort_keys = sort_keys

    def serialize(self, obj) -> bytes:
        return cbor.dumps(obj, self._sort_keys)

    def deserialize(self, payload: bytes):
        return cbor.loads(payload)

    @property
    def mimetype(self):
        return 'application/cbor'



from typeguard import check_argument_types
from typing import Any, Dict

import yaml

from ..api import Serializer


class YAMLSerializer(Serializer):
    """
    Serializes objects to and from YAML format.

    .. warning:: This serializer is insecure in unsafe mode because it allows execution of
      arbitrary code when deserializing.

    .. seealso:: `PyYAML documentation <http://pyyaml.org/wiki/PyYAMLDocumentation>`_

    :param safe: ``True`` to (de)serialize in safe mode, ``False`` to enable extended tags
    :param dumper_options: a dictionary of keyword arguments given to ``yaml.dump()``
    """

    __slots__ = '_dumper_class', '_loader_class', '_dumper_options'

    def __init__(self, safe: bool=True, dumper_options: Dict[str, Any]=None):
        assert check_argument_types()
        self._dumper_class = yaml.SafeDumper if safe else yaml.Dumper
        self._loader_class = yaml.SafeLoader if safe else yaml.Loader
        self._dumper_options = dumper_options or {}
        self._dumper_options.setdefault('encoding', 'utf-8')

    def serialize(self, obj) -> bytes:
        return yaml.dump(obj, Dumper=self._dumper_class, **self._dumper_options)

    def deserialize(self, payload: bytes):
        return yaml.load(payload, Loader=self._loader_class)

    @property
    def mimetype(self):
        return 'text/yaml'

    @property
    def safe(self):
        """Returns ``True`` if the safe mode is being used with (de)serialization."""
        return self._dumper_class is yaml.SafeDumper



import re

import pytest

from asphalt.serialization.serializers.cbor import CBORSerializer
from asphalt.serialization.serializers.json import JSONSerializer
from asphalt.serialization.serializers.msgpack import MsgpackSerializer
from asphalt.serialization.serializers.pickle import PickleSerializer
from asphalt.serialization.serializers.yaml import YAMLSerializer


@pytest.fixture(params=[
    CBORSerializer,
    JSONSerializer,
    MsgpackSerializer,
    PickleSerializer,
    YAMLSerializer
], ids=['cbor', 'json', 'msgpack', 'pickle', 'yaml'])
def serializer(request):
    return request.param()


@pytest.mark.parametrize('input', [
    'åäö',
    -8,
    5.06,
    [1, 'test', 1.23],
    {'x': 'foo', 'bar': 'baz'}
], ids=['str', 'int', 'float', 'list', 'dict'])
def test_basic_types_roundtrip(serializer, input):
    output = serializer.serialize(input)
    assert isinstance(output, bytes)

    deserialized = serializer.deserialize(output)
    assert deserialized == input


@pytest.mark.parametrize('serializer', [PickleSerializer(), YAMLSerializer()],
                         ids=['pickle', 'yaml'])
def test_circular_reference(serializer):
    a = {'foo': 1}
    b = {'a': a}
    a['b'] = b

    output = serializer.serialize(a)
    assert isinstance(output, bytes)

    other_a = serializer.deserialize(output)
    assert other_a['foo'] == 1
    other_b = other_a['b']
    assert other_b['a'] is other_a


def test_mime_types(serializer):
    assert re.match('[a-z]+/[a-z]+', serializer.mimetype)


@pytest.mark.parametrize('safe', [True, False], ids=['safe', 'unsafe'])
def test_yaml_safe_attribute(safe):
    serializer = YAMLSerializer(safe=safe)
    assert serializer.safe is safe



from typing import Dict, Any
from json.decoder import JSONDecoder
from json.encoder import JSONEncoder

from typeguard import check_argument_types
from asphalt.core.util import resolve_reference

from ..api import Serializer


class JSONSerializer(Serializer):
    """
    Serializes objects using JSON (JavaScript Object Notation).

    See the :mod:`json` module documentation in the standard library for more information on
    available options.

    Certain options can resolve references to objects:

    * ``encoder_options['default']``
    * ``decoder_options['object_hook']``
    * ``decoder_options['object_pairs_hook']``

    :param encoder_options: keyword arguments passed to :class:`~json.JSONEncoder`
    :param decoder_options: keyword arguments passed to :class:`~json.JSONDecoder`
    :param encoding: the text encoding to use for converting to and from bytes
    """

    __slots__ = 'encoding', '_encoder', '_decoder'

    def __init__(self, encoder_options: Dict[str, Any]=None, decoder_options: Dict[str, Any]=None,
                 encoding: str='utf-8'):
        assert check_argument_types()
        self.encoding = encoding
        encoder_options = encoder_options or {}
        encoder_options['default'] = resolve_reference(encoder_options.get('default'))
        self._encoder = JSONEncoder(**encoder_options)

        decoder_options = decoder_options or {}
        decoder_options['object_hook'] = resolve_reference(decoder_options.get('object_hook'))
        decoder_options['object_pairs_hook'] = resolve_reference(
            decoder_options.get('object_pairs_hook'))
        self._decoder = JSONDecoder(**decoder_options)

    def serialize(self, obj) -> bytes:
        return self._encoder.encode(obj).encode(self.encoding)

    def deserialize(self, payload: bytes):
        payload = payload.decode(self.encoding)
        return self._decoder.decode(payload)

    @property
    def mimetype(self):
        return 'application/json'



import logging
from typing import Dict, Any

from typeguard import check_argument_types

from asphalt.core.component import Component
from asphalt.core.context import Context
from asphalt.core.util import PluginContainer, merge_config
from asphalt.serialization.api import Serializer

serializer_backends = PluginContainer('asphalt.serialization.serializers', Serializer)

logger = logging.getLogger(__name__)


class SerializationComponent(Component):
    """
    Publishes one or more :class:`~asphalt.serialization.api.Serializer` resources.

    If more than one serializer is to be configured, provide a ``serializers`` argument as a
    dictionary where the key is the resource name and the value is a dictionary of keyword
    arguments to :meth:`configure_serializer`. Otherwise, directly pass those keyword arguments to
    the component constructor itself.

    If ``serializers`` is defined, any extra keyword arguments are used as default values for
    :meth:`configure_serializer` for all serializers (:func:`~asphalt.core.util.merge_config` is
    used to merge the per-serializer arguments with the defaults). Otherwise, a single serializer
    is created based on the provided default arguments, with ``context_attr`` defaulting to
    ``serializer``.

    :param serializers: a dictionary of resource name ⭢ :meth:`configure_serializer` arguments
    :param default_serializer_args: default values for omitted :meth:`configure_serializer`
        arguments
    """

    def __init__(self, serializers: Dict[str, Dict[str, Any]] = None, **default_serializer_args):
        assert check_argument_types()
        if not serializers:
            default_serializer_args.setdefault('context_attr', 'serializer')
            serializers = {'default': default_serializer_args}

        self.serializers = []
        for resource_name, config in serializers.items():
            config = merge_config(default_serializer_args, config or {})
            config.setdefault('backend', resource_name)
            config.setdefault('context_attr', resource_name)
            context_attr, serializer = self.configure_serializer(**config)
            self.serializers.append((resource_name, context_attr, serializer))

    @classmethod
    def configure_serializer(cls, context_attr: str, backend: str, **backend_args):
        """
        Configure a serializer.

        :param context_attr: context attribute of the serializer (if omitted, the resource name
            will be used instead)
        :param backend: entry point name used to look up the backend class (if omitted, the
            resource name will be used instead)
        :param backend_args: keyword arguments passed to the constructor of the backend class

        """
        assert check_argument_types()
        return context_attr, serializer_backends.create_object(backend, **backend_args)

    async def start(self, ctx: Context):
        for resource_name, context_attr, serializer in self.serializers:
            ctx.publish_resource(serializer, resource_name, context_attr, types=Serializer)
            logger.info('Configured serializer (%s / ctx.%s; type=%s)', resource_name,
                        context_attr, serializer.mimetype)



from abc import ABCMeta, abstractmethod


class Serializer(metaclass=ABCMeta):
    """
    This abstract class defines the serializer API.

    Each serializer is required to support the serialization of the following Python types,
    at minimum:

    * :class:`str`
    * :class:`int`
    * :class:`float`
    * :class:`list`
    * :class:`dict` (with ``str`` keys)

    A subclass may support a wider range of types, along with hooks to provide serialization
    support for custom types.
    """

    __slots__ = ()

    @abstractmethod
    def serialize(self, obj) -> bytes:
        """Serialize a Python object into bytes."""

    @abstractmethod
    def deserialize(self, payload: bytes):
        """Deserialize bytes into a Python object."""

    @property
    @abstractmethod
    def mimetype(self) -> str:
        """Return the MIME type for this serialization format."""



from typing import Dict, Any

from typeguard import check_argument_types

from ..api import Serializer

try:
    from msgpack import packb, unpackb
except ImportError:  # pragma: no cover
    raise ImportError('msgpack-python is missing -- install '
                      'asphalt-serialization[msgpack] to fix this')


class MsgpackSerializer(Serializer):
    """
    Serializes objects using the msgpack library.

    The following defaults are passed to packer/unpacker and can be overridden by setting values
    for the options explicitly:

    * ``use_bin_type=True`` (packer)
    * ``encoding='utf-8'`` (unpacker)

    To use this serializer backend, the ``msgpack-python`` library must be installed.
    A convenient way to do this is to install ``asphalt-serialization`` with the ``msgpack``
    extra:

    .. code-block:: shell

        $ pip install asphalt-serialization[msgpack]

    .. seealso:: `Msgpack web site <http://msgpack.org/>`_

    :param packer_options: keyword arguments passed to :func:`msgpack.packb`
    :param unpacker_options: keyword arguments passed to :func:`msgpack.unpackb`
    """

    __slots__ = '_packer_options', '_unpacker_options'

    def __init__(self, packer_options: Dict[str, Any]=None, unpacker_options: Dict[str, Any]=None):
        assert check_argument_types()
        packer_options = packer_options or {}
        packer_options.setdefault('use_bin_type', True)
        self._packer_options = packer_options

        unpacker_options = unpacker_options or {}
        unpacker_options.setdefault('encoding', 'utf-8')
        self._unpacker_options = unpacker_options

    def serialize(self, obj) -> bytes:
        return packb(obj, **self._packer_options)

    def deserialize(self, payload: bytes):
        return unpackb(payload, **self._unpacker_options)

    @property
    def mimetype(self):
        return 'application/x-msgpack'



from pathlib import Path

from setuptools import setup

setup(
    name='asphalt-serialization',
    use_scm_version={
        'version_scheme': 'post-release',
        'local_scheme': 'dirty-tag'
    },
    description='Serialization component for the Asphalt framework',
    long_description=Path(__file__).parent.joinpath('README.rst').read_text('utf-8'),
    author='Alex Grönholm',
    author_email='alex.gronholm@nextday.fi',
    url='https://github.com/asphalt-framework/asphalt-serialization',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5'
    ],
    license='Apache License 2.0',
    zip_safe=False,
    packages=[
        'asphalt.serialization',
        'asphalt.serialization.serializers'
    ],
    setup_requires=[
        'setuptools_scm >= 1.7.0'
    ],
    install_requires=[
        'asphalt ~= 2.0'
    ],
    extras_require={
        'msgpack': 'msgpack-python >= 0.4.6',
        'cbor': 'cbor >= 1.0'
    },
    entry_points={
        'asphalt.components': [
            'serialization = asphalt.serialization.component:SerializationComponent'
        ],
        'asphalt.serialization.serializers': [
            'cbor = asphalt.serialization.serializers.cbor:CBORSerializer [cbor]',
            'json = asphalt.serialization.serializers.json:JSONSerializer',
            'msgpack = asphalt.serialization.serializers.msgpack:MsgpackSerializer [msgpack]',
            'pickle = asphalt.serialization.serializers.pickle:PickleSerializer',
            'yaml = asphalt.serialization.serializers.yaml:YAMLSerializer'
        ]
    }
)
